/*
Copyright (c) 2000-2011 by Nicolas Devillard.
MIT License

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

/*-------------------------------------------------------------------------*/
/**
   @file	dictionary.c
   @author	N. Devillard
   @brief	Implements a dictionary for string variables.

   This module implements a simple dictionary object, i.e. a list
   of string/string associations. This object is useful to store e.g.
   informations retrieved from a configuration file (ini files).
*/
/*--------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
                                                                Includes
 ---------------------------------------------------------------------------*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <atalk/compat.h>
#include <atalk/dictionary.h>

/** Maximum value size for integers and doubles. */
#define MAXVALSZ 1024

/** Minimal allocated number of entries in a dictionary */
#define DICTMINSZ 128

/** Invalid key token */
#define DICT_INVALID_KEY ((char *)-1)

/*---------------------------------------------------------------------------
                                                        Private functions
 ---------------------------------------------------------------------------*/

#define MAXKEYSIZE 1024
static char *makekey(const char *section, const char *entry)
{
    static char buf[MAXKEYSIZE];

    strlcpy(buf, section, MAXKEYSIZE);
    if (entry) {
        strlcat(buf, ":", MAXKEYSIZE);
        strlcat(buf, entry, MAXKEYSIZE);
    }

    return buf;
}

/* Doubles the allocated size associated to a pointer */
/* 'size' is the current allocated size. */
static void *mem_double(void *ptr, int size)
{
    void *newptr;

    newptr = calloc(2 * size, 1);
    if (newptr == NULL) {
        return NULL;
    }
    memcpy(newptr, ptr, size);
    free(ptr);
    return newptr;
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Duplicate a string
  @param    s String to duplicate
  @return   Pointer to a newly allocated string, to be freed with free()

  This is a replacement for strdup(). This implementation is provided
  for systems that do not have it.
 */
/*--------------------------------------------------------------------------*/
static char *xstrdup(const char *s)
{
    char *t;
    size_t len;
    if (!s)
        return NULL;

    len = strlen(s) + 1;
    t = (char *)malloc(len);
    if (t) {
        memcpy(t, s, len);
    }
    return t;
}

/*---------------------------------------------------------------------------
                                                        Function codes
 ---------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/
/**
  @brief	Compute the hash key for a string.
  @param	key		Character string to use for key.
  @return	1 unsigned int on at least 32 bits.

  This hash function has been taken from an Article in Dr Dobbs Journal.
  This is normally a collision-free function, distributing keys evenly.
  The key is stored anyway in the struct so that collision can be avoided
  by comparing the key itself in last resort.
 */
/*--------------------------------------------------------------------------*/
unsigned atalkdict_hash(const char *key)
{
    int len;
    unsigned hash;
    int i;

    len = strlen(key);
    for (hash = 0, i = 0; i < len; i++) {
        hash += (unsigned)key[i];
        hash += (hash << 10);
        hash ^= (hash >> 6);
    }
    hash += (hash << 3);
    hash ^= (hash >> 11);
    hash += (hash << 15);
    return hash;
}

/*-------------------------------------------------------------------------*/
/**
  @brief	Create a new dictionary object.
  @param	size	Optional initial size of the dictionary.
  @return	1 newly allocated dictionary objet.

  This function allocates a new dictionary object of given size and returns
  it. If you do not know in advance (roughly) the number of entries in the
  dictionary, give size=0.
 */
/*--------------------------------------------------------------------------*/
dictionary *atalkdict_new(size_t size)
{
    dictionary *d;

    /* If no size was specified, allocate space for DICTMINSZ */
    if (size < DICTMINSZ)
        size = DICTMINSZ;

    d = (dictionary *)calloc(1, sizeof *d);

    if (d) {
        d->size = size;
        d->val = (char **)calloc(size, sizeof *d->val);
        d->key = (char **)calloc(size, sizeof *d->key);
        d->hash = (unsigned *)calloc(size, sizeof *d->hash);
    }
    return d;
}

/*-------------------------------------------------------------------------*/
/**
  @brief	Delete a dictionary object
  @param	d	dictionary object to deallocate.
  @return	void

  Deallocate a dictionary object and all memory associated to it.
 */
/*--------------------------------------------------------------------------*/
void atalkdict_del(dictionary *d)
{
    int i;

    if (d == NULL)
        return;
    for (i = 0; i < d->size; i++) {
        if (d->key[i] != NULL)
            free(d->key[i]);
        if (d->val[i] != NULL)
            free(d->val[i]);
    }
    free(d->val);
    free(d->key);
    free(d->hash);
    free(d);
    return;
}

/*-------------------------------------------------------------------------*/
/**
  @brief	Get a value from a dictionary.
  @param	d		dictionary object to search.
  @param	key		Key to look for in the dictionary.
  @param    def     Default value to return if key not found.
  @return	1 pointer to internally allocated character string.

  This function locates a key in a dictionary and returns a pointer to its
  value, or the passed 'def' pointer if no such key can be found in
  dictionary. The returned character pointer points to data internal to the
  dictionary object, you should not try to free it or modify it.
 */
/*--------------------------------------------------------------------------*/
const char *atalkdict_get(const dictionary *d, const char *section, const char *key, const char *def)
{
    unsigned hash;
    int i;

    hash = atalkdict_hash(makekey(section, key));
    for (i = 0; i < d->size; i++) {
        if (d->key[i] == NULL)
            continue;
        /* Compare hash */
        if (hash == d->hash[i]) {
            /* Compare string, to avoid hash collisions */
            if (!strcmp(makekey(section, key), d->key[i])) {
                return d->val[i];
            }
        }
    }
    return def;
}

/*-------------------------------------------------------------------------*/
/**
  @brief    Set a value in a dictionary.
  @param    d       dictionary object to modify.
  @param    key     Key to modify or add.
  @param    val     Value to add.
  @return   int     0 if Ok, anything else otherwise

  If the given key is found in the dictionary, the associated value is
  replaced by the provided one. If the key cannot be found in the
  dictionary, it is added to it.

  It is Ok to provide a NULL value for val, but NULL values for the dictionary
  or the key are considered as errors: the function will return immediately
  in such a case.

  Notice that if you atalkdict_set a variable to NULL, a call to
  atalkdict_get will return a NULL value: the variable will be found, and
  its value (NULL) is returned. In other words, setting the variable
  content to NULL is equivalent to deleting the variable from the
  dictionary. It is not possible (in this implementation) to have a key in
  the dictionary without value.

  This function returns non-zero in case of failure.
 */
/*--------------------------------------------------------------------------*/
int atalkdict_set(dictionary *d, const char *section, const char *key, const char *val)
{
    int i;
    unsigned hash;

    if (d == NULL || section == NULL)
        return -1;

    /* Compute hash for this key */
    hash = atalkdict_hash(makekey(section, key));
    /* Find if value is already in dictionary */
    if (d->n > 0) {
        for (i = 0; i < d->size; i++) {
            if (d->key[i] == NULL)
                continue;
            if (hash == d->hash[i]) {                            /* Same hash value */
                if (!strcmp(makekey(section, key), d->key[i])) { /* Same key */
                    /* Found a value: modify and return */
                    if (d->val[i] != NULL)
                        free(d->val[i]);
                    d->val[i] = val ? xstrdup(val) : NULL;
                    /* Value has been modified: return */
                    return 0;
                }
            }
        }
    }
    /* Add a new value */
    /* See if dictionary needs to grow */
    if (d->n == d->size) {

        /* Reached maximum size: reallocate dictionary */
        d->val = (char **)mem_double(d->val, d->size * sizeof(char *));
        d->key = (char **)mem_double(d->key, d->size * sizeof(char *));
        d->hash = (unsigned int *)mem_double(d->hash, d->size * sizeof(unsigned));
        if ((d->val == NULL) || (d->key == NULL) || (d->hash == NULL)) {
            /* Cannot grow dictionary */
            return -1;
        }
        /* Double size */
        d->size *= 2;
    }

    /* Insert key in the first empty slot */
    for (i = 0; i < d->size; i++) {
        if (d->key[i] == NULL) {
            /* Add key here */
            break;
        }
    }
    /* Copy key */
    d->key[i] = xstrdup(makekey(section, key));
    d->val[i] = val ? xstrdup(val) : NULL;
    d->hash[i] = hash;
    d->n++;
    return 0;
}

/*-------------------------------------------------------------------------*/
/**
  @brief	Delete a key in a dictionary
  @param	d		dictionary object to modify.
  @param	key		Key to remove.
  @return   void

  This function deletes a key in a dictionary. Nothing is done if the
  key cannot be found.
 */
/*--------------------------------------------------------------------------*/
void atalkdict_unset(dictionary *d, const char *section, const char *key)
{
    unsigned hash;
    int i;

    if (key == NULL) {
        return;
    }

    hash = atalkdict_hash(makekey(section, key));
    for (i = 0; i < d->size; i++) {
        if (d->key[i] == NULL)
            continue;
        /* Compare hash */
        if (hash == d->hash[i]) {
            /* Compare string, to avoid hash collisions */
            if (!strcmp(makekey(section, key), d->key[i])) {
                /* Found key */
                break;
            }
        }
    }
    if (i >= d->size)
        /* Key not found */
        return;

    free(d->key[i]);
    d->key[i] = NULL;
    if (d->val[i] != NULL) {
        free(d->val[i]);
        d->val[i] = NULL;
    }
    d->hash[i] = 0;
    d->n--;
    return;
}

/*-------------------------------------------------------------------------*/
/**
  @brief	Dump a dictionary to an opened file pointer.
  @param	d	Dictionary to dump
  @param	f	Opened file pointer.
  @return	void

  Dumps a dictionary onto an opened file pointer. Key pairs are printed out
  as @c [Key]=[Value], one per line. It is Ok to provide stdout or stderr as
  output file pointers.
 */
/*--------------------------------------------------------------------------*/
void atalkdict_dump(dictionary *d, FILE *out)
{
    int i;

    if (d == NULL || out == NULL)
        return;
    if (d->n < 1) {
        fprintf(out, "empty dictionary\n");
        return;
    }
    for (i = 0; i < d->size; i++) {
        if (d->key[i]) {
            fprintf(out, "%20s\t[%s]\n",
                    d->key[i],
                    d->val[i] ? d->val[i] : "UNDEF");
        }
    }
    return;
}
